(***************************************************************************
 *   Copyright (C) 2007  Antoine Bodin <antoinexp@gmail.com>               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        * 
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************)
 
(** Set up the lexer function *)

{
exception Parsing_failure of string

type token =
    Kwd of string
  | Ident of string
  | Bool of bool
  | Disp of bool
  | Int of Num.num
  | Float of float
  | String of string
  | Help of string
  | Void
  | End
  | EOF 
  | Newline
}

let digit = ['0'-'9']
let alnum = ['A'-'Z''a'-'z']
let variable = alnum (alnum|digit|'_')*
let space =  [' ' '\t' '\r']

let operator = ("+"|"-"|"*"|"/"|"^")
let loperator = ("="|"<>"|"<"|">"|"<="|">="|"not"|"and"|"or")

let loop = ("while"|"for"|"from"|"to"|"do"|"od")
let cond = ("if"|"then"|"else"|"elif"|"fi")
let keywords = ("("|")"|":="|","|"->"|"["|"]")
let display = [':' ';']
let lvalue = ("true"|"false")

let kwd = operator|loperator|loop|keywords|cond

rule lexer = parse
  | '\n' { [< 'Newline; lexer lexbuf >] }
  | space { lexer lexbuf }

  | digit+ as n { [< 'Int (Num.num_of_string n); lexer lexbuf >] }
  | digit+'.'digit* as f { [< 'Float (float_of_string f); lexer lexbuf >] }
  | lvalue as s { [< 'Bool (bool_of_string s); lexer lexbuf >] }

  | "\"" ([^'\"']* as s)  "\"" { [< 'String s; lexer lexbuf >] }

  | "end" { [< 'End; lexer lexbuf >] }
  | "()" { [< 'Void; lexer lexbuf >] }
  | ".." { [< 'Kwd ".."; lexer lexbuf >]}
  | (digit+ as n) ".." { [< 'Int (Num.num_of_string n); 'Kwd ".."; lexer lexbuf >]} (* necessary due to float numbers...*)
  | kwd as op { [< 'Kwd op; lexer lexbuf >] }

  | "\'" ([^'\'']+ as id)  "\'" { [< 'Ident id; lexer lexbuf >] }
  | variable as id { [< 'Ident id; lexer lexbuf >] }

  | display as c { [< 'Disp (c=';'); lexer lexbuf >] }
  | '?' ([^'\n']* as s) { [< 'Help s; lexer lexbuf >]}

  | "/*" { comments 0 lexbuf }
  | eof { [< 'EOF >] }
  | _ as c  { raise (Parsing_failure ("illegal character ("^(Char.escaped c)^")")) }

and comments lvl = parse
  | "/*" { comments (lvl+1) lexbuf}
  | "*/" { (if lvl=0 then lexer else comments (lvl-1)) lexbuf}
  | _ { comments lvl lexbuf }

{
  (*let test s = lexer (Lexing.from_string s)*)
}

