(***************************************************************************
 *   Copyright (C) 2007  Antoine Bodin <antoinexp@gmail.com>               *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        * 
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************)
 
(** The lexer of matl *)


open Lex


exception Syntax_error of string


let end_requested = ref false


let reset () = end_requested := true


let operator_of_string (op,i,j) =
  match op with
      "+" -> Core.Op (Core.Add,i,j)
    | "-" -> Core.Op (Core.Sub,i,j)
    | "*" -> Core.Op (Core.Mult,i,j)
    | "/" -> Core.Op (Core.Div,i,j)
    | "^" -> Core.Op (Core.Pow,i,j)
    | ".." -> Core.Op (Core.Interval,i,j)
    | "=" -> Core.LOp (Core.Eq,i,j)
    | "<>" -> Core.LOp (Core.Neq,i,j)
    | "<" -> Core.LOp (Core.Low,i,j)
    | ">" -> Core.LOp (Core.Gre,i,j)
    | "<=" -> Core.LOp (Core.Leq,i,j)
    | ">=" -> Core.LOp (Core.Geq,i,j)
    | "not" -> Core.LOp (Core.Lnot,i,j)
    | "and" -> Core.LOp (Core.Land,i,j)
    | "or" -> Core.LOp (Core.Lor,i,j)
    | _ -> failwith "unbound operator"


let makeinst b inst = {Core.verbose=b; Core.inst=inst}


let rec parse_operator opl func =
  let rec p e1 =
    parser 
	[< 'Kwd op when List.mem op opl; e2=(parse_operator opl func) ?? 
	  "right-side of the '"^op^"' operator forgotten?" >] -> 
	  operator_of_string (op,e1,e2)
      | [< 'Newline; expr >] -> p e1 expr (* usefull or not ? *)
      | [<>] -> e1
  in (parser [< e1=func; expr>] -> p e1 expr)


let disperr = "';' or ':' forgotten?"


let funerr = "a function must be bound to a variable, or a sequence of variables"


let rec parseproc stream = 
  let rec f = 
    parser
	[< 'Ident name; next >] -> name::f next
      | [< '(Kwd "," | Newline); next >] -> f next
      | [<>] -> []
  in match stream with parser
      [< 'Kwd "("; vl=f; 'Kwd ")"; seql=parseinst true; 'End; 'Ident "proc"; >] -> 
	Core.Proc (vl,seql)


and parseexpr expr =
  let rec seq =
    let rec p e1 =
      parser
	  [< 'Kwd ","; e2=seq >] ->
	    let gseq = function Core.Seq sq -> sq | e -> [e] in
	      Core.Seq ((gseq e1) @ (gseq e2))
	| [<>] -> e1
    in (parser [<e=lop; expr>] -> p e expr)
  and lop expr = parse_operator ["="; "<>"; "<"; ">"; "<="; ">="; "not"; "and"; "or"] op_under expr
  and op_under expr = parse_operator ["+"; "-"] op_over expr
  and op_over expr = parse_operator ["*"; "/"; "^"; ".."] fun_decl expr
  and fun_decl expr =
    let rec p e1 =
      parser
	| [< 'Void; expr >] -> p (Core.Call (e1,Core.Void)) expr
	| [< 'Kwd "("; e=seq; 'Kwd ")"; expr >] -> p (Core.Call (e1,e)) expr   (* todo: check if it works fine *)
	| [< 'Kwd "["; e=seq; 'Kwd "]"; expr >] -> p (Core.Call (Core.Var "get_item",
								Core.Seq [e1;e])) expr
	| [< 'Kwd "->"; e2=seq >] ->
	    let t =
	      match e1 with
		  Core.Var v -> [v]
		| Core.Void -> [] (*TODO *)
		| Core.Seq vl -> 
		    List.map (function Core.Var v -> v | _ -> raise (Syntax_error funerr)) vl;
		| _ -> raise (Syntax_error funerr);
	    in Core.Fun (t,e2)
	| [<>] -> e1
    in (parser [<e=value; stream>] -> p e stream) expr
  and value expr = 
    match expr with parser
	[< 'Int i >] -> Core.Int i
      | [< 'Bool b >] -> Core.Bool b
      | [< 'String s >] -> Core.String s
      |	[< 'Float i >] -> Core.Float i
      | [< 'Ident name; stream >] -> (match name with 
	    "proc" -> parseproc stream 
	  | _ -> Core.Var name)
      | [< 'Void >] -> Core.Void
      | [< 'Kwd "("; e=seq ?? "not an expression" ; 'Kwd ")" ?? "')' forgotten?">] -> e
      | [< 'Newline; expr >] -> value expr
  in seq expr


and parseinst depth inst =
  let parsedef e =
    parser
	[< 'Kwd ":="; expr=parseexpr>] -> 
	  (match e with
	      Core.Var name -> Core.Def (name,expr)
	    | _ -> raise (Syntax_error "wrong variable name"))
      | [< >] -> Core.Expr e
  in let rec parsecond =
    parser
	[< 'Kwd "elif"; cond=parseexpr; 'Kwd "then"; seql=parse true; next >] -> 
	  let seqlelse,b = parsecond next in
	    (makeinst b (Core.If (cond,seql,seqlelse))::[],b)
      | [< 'Kwd "else"; seql=parse true; 'Kwd "fi"; 'Disp b ?? disperr>] -> 
	  (seql,b)
      | [< 'Kwd "fi"; 'Disp b ?? disperr>] -> ([],b)
  and parse depth next =
    match next with parser
      | [< 'Newline; next >] -> if not depth then [] else parse true next
      | [< 'Ident "restart"; 'Disp b; next >] -> makeinst b Core.Restart::parse depth next
      | [< 'Ident "with"; 'Ident name; 'Disp b; next >] -> 
	  makeinst b (Core.Include name)::parse depth next
      | [< 'Kwd "while"; cond=parseexpr; 'Kwd "do"; seql=parse true; 'Kwd "od"; 'Disp b ?? disperr; next >] ->
	  makeinst b (Core.While (cond,seql))::parse depth next
      | [< 'Kwd "for"; 'Ident var; 'Kwd "from"; i=parseexpr; 'Kwd "to"; j=parseexpr; 'Kwd "do"; e=parse true; 'Kwd "od"; 'Disp b ?? disperr; next >] -> 
	  makeinst b (Core.For (var,i,j,e))::parse depth next
      | [< 'Kwd "if"; cond=parseexpr; 'Kwd "then"; seql=parse true; next >] ->
	  let seqlelse,b = parsecond next in
	    makeinst b (Core.If (cond,seql,seqlelse))::parse depth next
      | [< 'Help s; next >] -> Printf.printf "you are requesting for help: \"%s\" - Sorry, there is no help implemented yet, so rtfm or stfw.\n" s; parse depth next
      | [< e=parseexpr; inst=parsedef e; 'Disp b ?? disperr; next >] -> 
	  makeinst b inst::parse depth next
      | [< 'EOF >] -> end_requested := true; []
      | [<>] -> if depth then [] else (raise (Syntax_error "unexpected symbol"))
  in parse depth inst


let parseseql = parseinst false


let stream_from_channel inc = [< lexer (Lexing.from_channel inc) >]
    
